import numpy as np
import pandas as pd
import os
from sklearn.model_selection import train_test_split
from ampligraph.latent_features import DistMult, ComplEx, HolE, save_model, restore_model
from ampligraph.evaluation import train_test_split_no_unseen, evaluate_performance, mrr_score, hits_at_n_score, select_best_model_ranking
import pyiwn
from itertools import combinations
import pickle5 as pickle
import networkx as nx
from scipy.special import expit
import collections

print("## LINE 1 ##")

beng_iwn = pyiwn.IndoWordNet(lang = pyiwn.Language.BENGALI)
all_syns = beng_iwn.all_synsets()

print("## LINE 2 ##")

edges = pd.read_csv("../data/BengIwnData/edges.csv", header = None, delimiter = '\t', names = ['ent1','ent2', 'rel'])

print("## LINE 3 ##")

X = pd.DataFrame()
X['ent1'] = edges['ent1']
X['rel'] = edges['rel']
X['ent2'] = edges['ent2']
print("The dataframe X: \n", X.head())
X = X.astype(str)
#print("Value counts for relation:")
#print(X['rel'].value_counts())
#print(len(X['rel'].value_counts()))

print("## LINE 4 ##")

print("Changing the edge structure from (synset -> synset) to (word -> word)")

# converting synsets list into a hashmap
syn_dict = dict()
for val in all_syns:
  syn_dict[val.synset_id()] = val
  
def generate_triples(tuple_list, rel_id):
  triple_list = []
  for tup in tuple_list:
    triple = (tup[0], rel_id, tup[1])
    triple_list.append(triple)
  return triple_list

def generate_allcombs(list1, list2):
  res = []
  for val1 in list1:
    for val2 in list2:
      tup = (val1, val2)
      res.append(tup)
  return res
words_set = set()
word_edges = set()
visited_syns = set()
X = X.astype(int)
for i, row in X.iterrows():
  syn1_id = row['ent1']
  syn1 = syn_dict[int(syn1_id)]
  word_list1 = syn1.lemma_names()
  rel = row['rel']
  syn2_id = row['ent2']
  syn2 = syn_dict[int(syn2_id)]
  word_list2 = syn2.lemma_names()

  if(syn1_id not in visited_syns):
    # generating tuples for synonymy relation - id = 31
    synonymy_combs1 = combinations(word_list1, 2)
    synonymy_triples1 = generate_triples(synonymy_combs1, '31') # add to set
    for i, val in enumerate(synonymy_triples1):
      word_edges.add(val)
    visited_syns.add(syn1_id)

  if(syn2_id not in visited_syns):
    synonymy_combs2 = combinations(word_list2, 2)
    synonymy_triples2 = generate_triples(synonymy_combs2, '31') # add to set
    for i, val in enumerate(synonymy_triples2):
      word_edges.add(val)
    visited_syns.add(syn2_id)

  # enumerating the same relation for all words in both synset lists
  word_combinations = generate_allcombs(word_list1, word_list2)
  word_triples = generate_triples(word_combinations, str(rel)) # add to set

  for i, val in enumerate(word_triples):
    word_edges.add(val)

print("## LINE 6 ##")

print(f"Number of edges with words as nodes = {len(word_edges)}")
word_edgelist = list(word_edges)
word_dict = collections.defaultdict(int)
for i,val in enumerate(word_edgelist):
    word_dict[val] = 1
word_edgearr = np.asarray(word_edgelist)
#print(word_edgearr[:5])
# splitting into train, test and validation
X_train, X_test = train_test_split_no_unseen(word_edgearr, seed = 1, test_size = 20000)
print("Training set = ", X_train.shape)
print("Test set = ", X_test.shape)

positive_filter = word_edgearr

def evaluate(model):
  test_ranks = evaluate_performance(X_test, model=model, filter_triples = positive_filter)
  mrr_test = mrr_score(test_ranks)
  hits_at_1_test = hits_at_n_score(test_ranks, 1)
  hits_at_3_test = hits_at_n_score(test_ranks, 3)
  hits_at_10_test = hits_at_n_score(test_ranks, 10)

  print("Results:")
  print(f"Test mrr = {mrr_test}")
  print(f"Test hits_at_1 score = {hits_at_1_test}")
  print(f"Test hits_at_3 score = {hits_at_3_test}")
  print(f"Test hits_at_10 score = {hits_at_10_test}")
   

def generate_all_rel(u,v):
    result = []
    for rel_int in range(1,32):
        rel_triple = (u, str(rel_int), v)
        result.append(rel_triple)
    return result

# reading subgraph
subgraph_file = open("./subgraph3.pkl", "rb")
dt = pickle.load(subgraph_file)

dt_edgelist = []
count = 0
for u,v in dt.edges():
    dt_edgelist.extend(generate_all_rel(u,v))

dt_unseen = []
for i,val in enumerate(dt_edgelist):
    if word_dict[val] != 1:
        dt_unseen.append(val)

print(f"No. of unseen edges = {len(dt_unseen)}")

print(f"Taking up second half of unseen edges:")

rem = -1 * (len(dt_unseen) - (len(dt_unseen)//2))
dt_slice = dt_unseen[rem:]
dt_edgearr = np.asarray(dt_slice)

print(f"Number of unseen edges to work on = {len(dt_edgearr)}")

unseen_filter = np.array(list({tuple(i) for i in np.vstack((positive_filter, dt_edgearr))}))

def get_unseen_scores(model, name):
    ranks_unseen = evaluate_performance(
            dt_edgearr,
            model = model,
            filter_triples = unseen_filter,
            filter_unseen=True
            )
    scores = model.predict(dt_edgearr)
    # converting to probabilities
    probs = expit(scores)
    score_df = pd.DataFrame(list(zip([' '.join(x) for x in dt_edgearr],
        ranks_unseen,
        np.squeeze(scores),
        np.squeeze(probs))),
        columns = ['triple', 'rank', 'score', 'prob']).sort_values("score", ascending = False)
    #score_df.to_csv(name)
    return score_df

#distmult_model = DistMult(batches_count= 100, epochs=10, k= 100, eta= 50, optimizer= 'adagrad', loss= 'nll', regularizer= 'LP', verbose= False, optimizer_params= {'lr': 0.1}, regularizer_params= {'p': 2, 'lambda': 0.001}, embedding_model_params= {'negative_corruption_entities': 'all'})

print("DistMult Model:")
distmult_model = restore_model(model_name_path = "distmult.pkl")
print("Model restored")
#evaluate(distmult_model)
dist_scores = get_unseen_scores(distmult_model, "distmult1.csv")

#complex_model = ComplEx(batches_count= 100, epochs=10, k= 100, eta= 50, optimizer= 'adagrad', loss= 'pairwise', regularizer= 'LP', verbose= False, optimizer_params= {'lr': 0.01}, regularizer_params= {'p': 2, 'lambda': 0.0001}, loss_params= {'margin': 0.1}, embedding_model_params= {'negative_corruption_entities': 'all'})

print("ComplEx Model:")
complex_model = restore_model(model_name_path = "complex.pkl")
print("Model restored")
#evaluate(complex_model)
comp_scores = get_unseen_scores(complex_model, "complex1.csv")

#hole_model = HolE(batches_count= 100, epochs=10, k= 100, eta= 50, optimizer= 'adagrad', loss= 'pairwise', regularizer= None, verbose= False, optimizer_params= {'lr': 0.1}, loss_params= {'margin': 0.5}, embedding_model_params= {'negative_corruption_entities': 'all'})

print("HolE Model:")
hole_model = restore_model(model_name_path = "hole.pkl")
print("Model restored")
#evaluate(hole_model)
hole_scores = get_unseen_scores(hole_model, "hole1.csv")

dist_pos = dist_scores.loc[dist['score'] > 0]
print(f"No. of triples predicted to have positive scores with distmult = {len(dist_pos)}")

hole_pos = hole_scores.loc[hole['score'] > 0]
print(f"No. of triples predicted to have positive scores with HolE = {len(hole_pos)}")

comp_pos = comp_scores.loc[comp['score'] > 0]
print(f"No. of triples predicted to have positive scores with ComplEx = {len(comp_pos)}")

# checking the count of how many are not already present

def check_unseen(df, res):
    for i, row in df.iterrows():
        cur = row['triple']
        triple = tuple(cur.split(' '))
        if triple not in word_edgelist:
            res.append(triple)
    return res

dist_list = []
comp_list = []
hole_list = []

#dist_list = check_unseen(dist_pos, dist_list)
#comp_list = check_unseen(comp_pos, comp_list)
#hole_list = check_unseen(hole_pos, hole_list)

print("Distmult :")
print(dist_pos.head())
#print(f"No. of unseen triples given a positive score: {len(dist_list)}")

print("ComplEx: ")
print(comp_pos.head())
#print(f"No. of unseen triples given a positive score: {len(comp_list)}")

print("HolE: ")
print(hole_pos.head())
#print(f"No. of unseen triples given a positive score: {len(hole_list)}")













































































